import torch
import torchvision
import torchvision.transforms as transforms
import numpy as np

class CIFAR10Dataset(torch.utils.data.Dataset):
  def __init__(self, transform, data, label):
    # 调用父类的构造函数
    super(DIV2KDataset, self).__init__()
    self.transform = transform   # 设置属性 transform
    self.images = data            # 假设data的shape为
                                    # (图片数, 32, 32, 3)
                                    # 数据类型是 np.uint8
                                    # 值域为 [0,255]
    self.labels = label           # 假设label的shape为 (图片数,)
                                    # PyTorch 会在计算交叉熵
                                    # 自动转为 One-Hot 编码 
  
  # 相当于重载 [] 运算符
  def __getitem__(self, idx):
    img = self.images[idx]
    img = self.transform(img)
    label = self.labels[idx]
    return img, label
  
  # 实现该方法之后，对于该类的一个对象，可以用 len(obj) 来获取长度
  def __len__(self):
    return len(self.images)

# 定义 transform：包括两个顺序步骤
# 1，把numpy数组转化为pytorch张量
# 2，归一化到 [-0.5, 0.5]，有利于 ReLU 函数处理
transform = transforms.Compose(  
    [transforms.ToTensor(), 
     transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))]
    )

# 假设数据已提前准备好
train_data = np.load('train_data.npy')
train_label = np.load('train_label.npy')
test_data = np.load('test_data.npy')
test_label = np.load('test_label.npy')

# trainset是一个CIFAR10Dataset实例，可以用下标索引
# 下标索引会返回一个sample的data和label
Trainset = CIFAR10Dataset(
                  transform=transform, 
                  data=train_data,
                  label=train_label
             )
# PyTorch提供的dataloader可以方便地控制batchsize和shuffle
# 以及提供了异步接口
# 如果出现异步问题，设置num_workers=0
trainloader = torch.utils.data.DataLoader(
                   trainset, batch_size=4,
                   shuffle=True, num_workers=2
                )

testset = CIFAR10Dataset(
               transform=transform, data=test_data, label=test_label
           )

import torch.nn as nn
import torch.nn.functional as F

# 继承一个nn.Module，实现了构造函数和forward方法，就是一个网络模型
class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        # 2维卷积，输入通道3，输出通道6，卷积核大小5x5
        # 还有其它参数可以设置 (stride, padding)
        self.conv1 = nn.Conv2d(3, 6, 5)        
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        # fc fully connected，全连接层
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        # view函数进行reshape操作
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


# 实例化一个神经网络
net = Net()
# net = Net().cuda() 取消注释则将神经网络模型放至GPU上
criterion = nn.CrossEntropyLoss()
# Stochastic Gradient Descent
optimizer = optim.SGD(net.parameters(), lr=0.001, momentum=0.9)
for epoch in range(10):
	for i, data in enumerate(trainloader, 0):
		# 获取输入数据，trainloader 每次会产生一个元组
    	# 也就是 Dataset 返回的 data 和 label
		inputs, labels = data
    	# 但是，因为用DataLoader包装过，因此PyTorch会自动帮忙组成batch
    	# data.shape: (batchsize, 3, 32, 32)
    	# label.shape: (batchsize, 10)
    	# 如果使用GPU，要通过 inputs = inputs.cuda() 把数据放至GPU上
    	# 同样地，设置 labels = labels.cuda()
		optimizer.zero_grad()
		# 前向传播
    	outputs = net(inputs)
    	# 计算 loss
		loss = criterion(outputs, labels)
    	# 反向传播
		loss.backward()
		# 更新
    	optimizer.step()


    # 在每个epoch存储模型
    PATH = './cifar_net.pth'
    torch.save(net.state_dict(), PATH)

net = Net()
# 加载之前训好的模型参数
net.load_state_dict(torch.load(PATH))
# 建立列表统计正确个数和总数
class_correct = list(0. for i in range(10))
class_total = list(0. for i in range(10))
# 以下部分不需要反向传播
with torch.no_grad():
    for data in testloader:
        images, labels = data
        outputs = net(images)
        # torch.max 执行在outputs的维度1上
        # 每次会输出两项，即max和argmax
        _, predicted = torch.max(outputs, 1)
        # 使用一个简便的方法统计正确个数
        c = (predicted == labels).squeeze()
        # 对于批中的每个样本的正确与否，按实际标签进行分别统计
        for i in range(batchsize):
            label = labels[i]
            class_correct[label] += c[i].item()
            class_total[label] += 1

    for i in range(10):
        print('Accuracy of %5s : %2d %%' % (
            classes[i], 100 * class_correct[i] / class_total[i]))
